' vim: set sts=4 ts=4 sw=4 et co=88 tw=87 ft=vb fo=tcqro:
Option Explicit

Imports System
Imports System.ComponentModel
Imports System.Runtime.InteropServices
Imports System.Threading
Imports VB = Microsoft.VisualBasic

Namespace Games
    Public Class NativeMethods
        Public Const STD_INPUT_HANDLE As Integer = -10
        Public Const KEY_EVENT As Short = &H0001
        Public Const CAPSLOCK_ON As Integer = &H0080
        Public Const ENHANCED_KEY As Integer = &H0100
        Public Const LEFT_ALT_PRESSED As Integer = &H0002
        Public Const LEFT_CTRL_PRESSED As Integer = &H0008
        Public Const NUMLOCK_ON As Integer = &H0020
        Public Const RIGHT_ALT_PRESSED As Integer = &H0001
        Public Const RIGHT_CTRL_PRESSED As Integer = &H0004
        Public Const SCROLLLOCK_ON As Integer = &H0040
        Public Const SHIFT_PRESSED As Integer = &H0010
        Public Shared INVALID_HANDLE_VALUE As IntPtr = New IntPtr(-1)
        Public Const WAIT_OBJECT_0 As Integer = CInt(&H00000000)
        Public Const WAIT_FAILED As Integer = CInt(&HFFFFFFFF)
        Public Const INFINITE As Integer = CInt(&HFFFFFFFF)

        Declare Unicode Function ReadConsoleInput Lib "kernel32.dll" Alias _
            "ReadConsoleInputW" _
            (ByVal hConsoleInput As IntPtr, _
            ByRef lpBuffer As INPUT_RECORD_KEY, _
            ByVal nLength As Integer, _
            ByRef lpNumberOfEventsRead As Integer) _
            As Integer

        Declare Unicode Function PeekConsoleInput Lib "kernel32.dll" Alias _
            "PeekConsoleInputW" _
            (ByVal hConsoleInput As IntPtr, _
            ByRef lpBuffer As INPUT_RECORD_KEY, _
            ByVal nLength As Integer, _
            ByRef lpNumberOfEventsRead As Integer) _
            As Integer

        Declare Auto Function GetStdHandle Lib "kernel32.dll" _
            (ByVal nStdHandle As Integer) _
            As IntPtr

        Declare Auto Function WaitForSingleObject Lib "kernel32.dll" _
            (ByVal hHandle As IntPtr, _
            ByVal dwMilliseconds As Integer) _
            As Integer

        <StructLayout(LayoutKind.Sequential)> _
        Public Structure INPUT_RECORD_KEY
            Public EventType As Short
            Public KeyEvent As KEY_EVENT_RECORD
        End Structure

        <StructLayout(LayoutKind.Sequential, CharSet := CharSet.Unicode)> _
        Public Structure KEY_EVENT_RECORD
            Public bKeyDown As Integer
            Public wRepeatCount As Short
            Public wVirtualKeyCode As Short
            Public wVirtualScanCode As Short
            Public uChar As Char ' marshal as a Unicode character (see CharSet)
            Public dwControlKeyState As Integer
        End Structure
    End Class

    Public Class GuessNum
        Private Const NumberLength As Integer = 4

        Public Sub New()
        End Sub

        Public Shared Sub Main()
            Dim obj As GuessNum = New GuessNum()

            VB.Randomize()
            obj.RunGame()
        End Sub

        Public Sub RunGame()
            WelcomePlayer()
            ShowConfig()
            DoGuessLoop()
            DrinkCoffee()
        End Sub

        Public Sub WelcomePlayer()
            Console.WriteLine("Welcome to Robbie's Cai Shuzi (guess number) game.")
            Console.WriteLine("The game is very simple: the computer generates a digit sequence.")
            Console.WriteLine("But it doesn't show it to you.  You guess it.")
        End Sub

        Public Sub ShowConfig()
            Console.WriteLine("The length of the digit sequence is {0}.", NumberLength)
        End Sub

        Public Sub DoGuessLoop()
            Dim gn As GnNumber
            Dim agAns As String ' again answer

            Do
                gn = New GnNumber(NumberLength)
                Do While GuessOnce(gn)
                Loop
                Console.Write("Do you want to play again [Y/N]? ")
                agAns = Console.ReadLine()
                If VB.LCase(agAns) <> "y" AndAlso VB.LCase(agAns) <> "yes" Then
                    Exit Do
                End If
            Loop
        End Sub

        Public Function GuessOnce(gn As GnNumber) As Boolean
            Dim line As String

            Console.Write("Your guess? ")
            line = Console.ReadLine()
            If line Is Nothing Then
                line = ""
            End If
            If line = "" Then
                Console.WriteLine("Empty input, quitting.")
                Return False
            End If
            gn.UserGuess = line
            gn.JudgeMatch()
            If gn.MatchACount = NumberLength Then
                Console.WriteLine("Bingo!  The number is: {0}.", gn.NumberS)
                return False
            Else
                Console.WriteLine("{0}A{1}B", gn.MatchACount, gn.MatchBCount)
                Return True
            End If
        End Function

        Public Sub DrinkCoffee()
            Dim pm As PacMan

            pm = New PacMan()
            Console.WriteLine("It's now time for a cup of coffee.")
            Console.WriteLine("Please wait for your coffee to be ready.")
            Console.WriteLine("Now you can type some characters for fun.")
            Thread.Sleep(10000)
            Console.WriteLine("Coffee ready to serve.  Enjoy!")
            Console.WriteLine("Hope you like this game and come back later.")
            pm.SignalEndAndWait()
        End Sub

        Public Sub GoPacman()
            Console.Read()
        End Sub
    End Class

    Public Class GnNumber
        Private myNumS As String
        Private myGuess As String
        Private myACount As Integer
        Private myBCount As Integer

        Public Sub New(length As Integer)
            myNumS = GenerateAsString(length)
        End Sub

        Public Sub JudgeMatch()
            Dim aCount As Integer
            Dim bCount As Integer
            Dim i As Integer

            aCount = 0
            ' Compute A Count
            For i = 0 To Math.Min(myGuess.Length - 1, myNumS.Length - 1)
                ' Exact match is A Count
                If myGuess.Chars(i) = myNumS.Chars(i) Then
                    aCount += 1
                End If
            Next
            bCount = ComputeBCount(myNumS, myGuess)
            myACount = aCount
            myBCount = bCount
        End Sub

        ''' <summary>
        ''' Generates a random number for guessing as a string
        ''' </summary>
        Private Shared Function GenerateAsString(length As Integer) As String
            Dim cands(10) As Char ' Candidates
            Dim rand As Integer
            Dim tmp As Char
            Dim i As Integer
            Dim result As String

            ' Put digits in an array
            For i = 0 To 10 - 1
                cands(i) = VB.Chr(VB.Asc("0") + i)
            Next
            ' Randomize the digits
            For i = 0 To 10 - 1
                rand = VB.Int(VB.Rnd() * (10 - i))
                tmp = cands(i)
                cands(i) = cands(10 - 1 - rand)
                cands(10 - 1 - rand) = tmp
            Next
            ' Get the first a few digits
            result = ""
            For i = 0 To length - 1
                result = result + cands(i)
            Next
            Return result
        End Function

        Public Shared Function ComputeBCount(numS As String, guess As String) As Integer
            Dim guessB As String
            Dim bCount As Integer
            Dim i As Integer
            Dim j As Integer

            guessB = ""
            ' Copy to guessB without A Count characters
            For i = 0 To Math.Min(guess.Length - 1, numS.Length - 1)
                ' Exact match is A Count
                If guess.Chars(i) <> numS.Chars(i) Then
                    guessB += guess.Chars(i)
                End If
            Next
            ' Compute B Count: number of characters of `guessB' that are in `numB'
            bCount = 0
            For i = 0 To guessB.Length - 1
                For j = 0 To numS.Length - 1
                    If numS.Chars(j) = guessB.Chars(i) Then
                        Exit For
                    End If
                Next
                If j < numS.Length Then
                    bCount += 1
                End If
            Next
            Return bCount
        End Function

        Public ReadOnly Property NumberS As String
            Get
                Return myNumS
            End Get
        End Property

        Public Property UserGuess As String
            Get
                Return myGuess
            End Get
            Set(Value As String)
                myGuess = Value
            End Set
        End Property

        Public ReadOnly Property MatchACount As Integer
            Get
                Return myACount
            End Get
        End Property

        Public ReadOnly Property MatchBCount As Integer
            Get
                Return myBCount
            End Get
        End Property
    End Class

    Public Class PacMan
        Private myThread As Thread
        Private myEndFlag As Boolean
        Private myEnded As Boolean

        Public Sub New()
            Me.EndFlag = False
            Me.Ended = False
            myThread = New Thread(AddressOf GoPacMan)
            myThread.Start()
        End Sub

        Public Sub GoPacMan()
            Dim stdIn As IntPtr
            Dim dwRet As Integer

            stdIn = NativeMethods.GetStdHandle(NativeMethods.STD_INPUT_HANDLE)
            If stdIn.Equals(NativeMethods.INVALID_HANDLE_VALUE) Then
                Throw New Win32Exception()
            End If
            Do While Not Me.EndFlag
                dwRet = NativeMethods.WaitForSingleObject(stdIn, 1000)
                If dwRet = NativeMethods.WAIT_FAILED Then
                    Throw New Win32Exception()
                End If
                AutoReadAndShowChar(stdIn)
            Loop
            Me.Ended = True
        End Sub

        Private Sub AutoReadAndShowChar(stdIn As IntPtr)
            ' NOTE: use "= 0" to compare Integer rather than "Not" (bit-wise)
            Dim bRet As Integer ' receives boolean return value;
            Dim buf As NativeMethods.INPUT_RECORD_KEY
            Dim dwRet As Integer

            bRet = NativeMethods.PeekConsoleInput(stdIn, buf, 1, dwRet)
            If bRet = 0 Then
                Throw New Win32Exception()
            End If
            If dwRet = 1 Then
                If buf.KeyEvent.bKeyDown <> 0 AndAlso Not HasControlKey(buf.KeyEvent. _
                dwControlKeyState) AndAlso VB.AscW(buf.KeyEvent.uChar) <> 0 Then
                    ShowChar(buf.KeyEvent.uChar)
                End If
                bRet = NativeMethods.ReadConsoleInput(stdIn, buf, 1, dwRet)
                If bRet = 0 Then
                    Throw New Win32Exception()
                End If
            ' Else do nothing
            End If
        End Sub

        Private Function HasControlKey(dwControlKeyState As Integer) As Boolean
            If dwControlKeyState AndAlso (NativeMethods.LEFT_ALT_PRESSED OrElse _
            NativeMethods.LEFT_CTRL_PRESSED OrElse NativeMethods.RIGHT_ALT_PRESSED _
            OrElse NativeMethods.RIGHT_CTRL_PRESSED) Then
                Return True
            End If
            Return False
        End Function

        Private Sub ShowChar(ch As Char)
            If VB.AscW(ch) = 13 Then
                Console.WriteLine()
            Else
                Console.Write(ch)
            End If
        End Sub

        Public Sub SignalEndAndWait()
            Me.EndFlag = True
            Do While Not Me.Ended
            Loop
        End Sub

        Public Property EndFlag As Boolean
            Get
                Dim result As Boolean

                SyncLock Me
                    result = myEndFlag
                End SyncLock
                Return result
            End Get
            Set(Value As Boolean)
                SyncLock Me
                    myEndFlag = Value
                End SyncLock
            End Set
        End Property

        Public Property Ended As Boolean
            Get
                Dim result As Boolean

                SyncLock Me
                    result = myEnded
                End SyncLock
                Return result
            End Get
            Set(Value As Boolean)
                SyncLock Me
                    myEnded = Value
                End SyncLock
            End Set
        End Property
    End Class
End Namespace
